"""
.. module:: test_problems
  :synopsis: Test problems for multi-modal and
             box-constrained global optimization
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
                 David Bindel <bindel@cornell.edu>

:Module: test_problems
:Author: David Eriksson <dme65@cornell.edu>
    David Bindel <bindel@cornell.edu>
"""

import random
from time import time
import numpy as np
from poap.controller import ProcessWorkerThread
import importlib

def validate(obj):
    """Routine for checking that an implementation of an objective function
    follows the standard. This method checks everything, but can't make
    sure that the objective function and constraint methods return values
    of the correct type since this would involve actually evaluating the
    objective function which isn't feasible when the evaluations are
    expensive. If some test fails, an exception is raised through assert.

    :param obj: Objective function
    """
    assert hasattr(obj, "dim"), \
        "Problem dimension required"
    assert hasattr(obj, "xlow"), \
        "Numpy array of lower bounds required"
    assert isinstance(obj.xlow, np.ndarray), \
        "Numpy array of lower bounds required"
    assert hasattr(obj, "xup"), \
        "Numpy array of upper bounds required"
    assert isinstance(obj.xup, np.ndarray), \
        "Numpy array of upper bounds required"
    assert hasattr(obj, "integer"), \
        "Integer variables must be specified"
    if len(obj.integer) > 0:
        assert isinstance(obj.integer, np.ndarray), \
            "Integer variables must be specified"
    else:
        assert isinstance(obj.integer, np.ndarray) or \
            isinstance(obj.integer, list), \
            "Integer variables must be specified"
    assert hasattr(obj, "continuous"), \
        "Continuous variables must be specified"
    if len(obj.continuous) > 0:
        assert isinstance(obj.continuous, np.ndarray), \
            "Continuous variables must be specified"
    else:
        assert isinstance(obj.continuous, np.ndarray) or \
            isinstance(obj.continuous, list), \
            "Continuous variables must be specified"
    if not hasattr(obj, "objfunction"):
        try:
            module = importlib.import_module(obj.__module__)
            objfun = getattr(module, "objfunction")
            if not issubclass(objfun, ProcessWorkerThread):
                raise AssertionError("Method 'objfunction' is not implemented")
        except:
            raise AssertionError("Method 'objfunction' is not implemented")

    # Check for logical errors
    assert isinstance(obj.dim, int) and obj.dim > 0, \
        "Problem dimension must be a positive integer."
    assert (len(obj.xlow) == obj.dim and
            len(obj.xup) == obj.dim), \
        "Incorrect size for xlow and xup"
    assert all(obj.xlow[i] < obj.xup[i] for i in range(obj.dim)), \
        "Lower bounds must be below upper bounds."
    if len(obj.integer) > 0:
        assert np.amax(obj.integer) < obj.dim and np.amin(obj.integer) >= 0, \
            "Integer variable index can't exceed " \
            "number of dimensions or be negative"
    if len(obj.continuous) > 0:
        assert np.amax(obj.continuous) < obj.dim and \
               np.amin(obj.continuous) >= 0, \
               "Continuous variable index can't exceed " \
               "number of dimensions or be negative"
    assert len(np.intersect1d(obj.continuous, obj.integer)) == 0, \
        "A variable can't be both an integer and continuous"
    assert len(obj.continuous)+len(obj.integer) == obj.dim, \
        "All variables must be either integer or continuous"

# ========================= 3-dimensional =======================


class Hartman3:
    """Hartman 3 function

    Details: http://www.sfu.ca/~ssurjano/hart3.html

    Global optimum: :math:`f(0.114614,0.555649,0.852547)=-3.86278`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=3):
        self.xlow = np.zeros(3)
        self.xup = np.ones(3)
        self.dim = 3
        self.info = "3-dimensional Hartman function \nGlobal optimum: " +\
                    "f(0.114614,0.555649,0.852547) = -3.86278"
        self.min = -3.86278
        self.integer = []
        self.continuous = np.arange(0, 3)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Hartman 3 function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        alpha = np.array([1, 1.2, 3, 3.2])
        A = np.matrix([[3.0, 10.0, 30.0], [0.1, 10.0, 35.0],
                       [3.0, 10.0, 30.0], [0.1, 10.0, 35.0]])
        P = np.matrix([[0.3689, 0.1170, 0.2673],
                       [0.4699, 0.4387, 0.747],
                       [0.1091, 0.8732, 0.5547],
                       [0.0381, 0.5743, 0.8828]])
        outer = 0
        for ii in range(4):
            inner = 0
            for jj in range(3):
                xj = x[jj]
                Aij = A[ii, jj]
                Pij = P[ii, jj]
                inner += Aij * ((xj-Pij) ** 2)
            outer += alpha[ii] * np.exp(-inner)
        return -outer

# ========================= n-dimensional =======================


class Rastrigin:
    """Rastrigin function

    .. math::
        f(x_1,\\ldots,x_n)=10n-\\sum_{i=1}^n (x_i^2 - 10 \\cos(2 \\pi x_i))

    subject to

    .. math::
        -5.12 \\leq x_i \\leq 5.12

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -5.12 * np.ones(dim)
        self.xup = 5.12 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Rastrigin function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Rastrigin function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return 10 * self.dim + sum(x**2 - 10 * np.cos(2 * np.pi * x))


class Ackley:
    """Ackley function

    .. math::
        f(x_1,\\ldots,x_n) = -20\\exp\\left( -0.2 \\sqrt{\\frac{1}{n} \
        \\sum_{j=1}^n x_j^2} \\right) -\\exp \\left( \\frac{1}{n} \
        \\sum{j=1}^n \\cos(2 \\pi x_j) \\right) + 20 - e

    subject to

    .. math::
        -15 \\leq x_i \\leq 20

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -15 * np.ones(dim)
        self.xup = 20 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Ackley function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Ackley function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        n = float(len(x))
        return -20.0 * np.exp(-0.2*np.sqrt(sum(x**2)/n)) - \
            np.exp(sum(np.cos(2.0*np.pi*x))/n) + 20 + np.exp(1)


class Michalewicz:
    """Michalewicz function

    """
    def __init__(self, dim=10):
        self.xlow = np.zeros(dim)
        self.xup = np.pi * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Michalewicz function \n" +\
                             "Global optimum: ??"
        self.min = np.NaN
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Michalewicz function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return -np.sum(np.sin(x) * (np.sin(((1+np.arange(self.dim))
                                            * x**2)/np.pi)) ** 20)


class Levy:
    def __init__(self, dim=10):
        self.xlow = -5 * np.ones(dim)
        self.xup = 5 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Levy function \n" +\
                             "Global optimum: ?"
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Levy function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        a = 20
        b = 0.2
        d = 50
        return -a * np.exp(-b * np.sqrt(np.sum(x ** 2)/d)) -\
            np.exp(np.sum(np.cos(2 * np.pi * x))/d)


class Griewank:
    """Griewank function

    .. math::
        f(x_1,\\ldots,x_n) = 1 + \\frac{1}{4000} \\sum_{j=1}^n x_j^2 - \
        \\prod_{j=1}^n \\cos \\left( \\frac{x_i}{\\sqrt{i}} \\right)

    subject to

    .. math::
        -512 \\leq x_i \\leq 512

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -512 * np.ones(dim)
        self.xup = 512 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Griewank function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Griewank function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 1
        for i, y in enumerate(x):
            total *= np.cos(y / np.sqrt(i+1))
        return 1.0 / 4000.0 * sum([y**2 for y in x]) - total + 1


class Rosenbrock:
    """Rosenbrock function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^{n-1} \
        \\left( 100(x_j^2-x_{j+1})^2 + (1-x_j)^2 \\right)

    subject to

    .. math::
        -2.048 \\leq x_i \\leq 2.048

    Global optimum: :math:`f(1,1,...,1)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -2.048 * np.ones(dim)
        self.xup = 2.048 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Rosenbrock function \n" +\
                             "Global optimum: f(1,1,...,1) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Rosenbrock function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0
        for i in range(len(x) - 1):
            total += 100 * (x[i] ** 2 - x[i+1]) ** 2 + (x[i] - 1) ** 2
        return total


class Schwefel:
    """Schwefel function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^{n} \
        \\left( -x_j \\sin(\\sqrt{|x_j|}) \\right) + 418.982997 n

    subject to

    .. math::
        -512 \\leq x_i \\leq 512

    Global optimum: :math:`f(420.968746,420.968746,...,420.968746)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -512 * np.ones(dim)
        self.xup = 512 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Schwefel function \n" +\
                             "Global optimum: f(420.968746,...,420.968746) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Schwefel function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return 418.9829 * self.dim - \
            sum([y * np.sin(np.sqrt(abs(y))) for y in x])


class Sphere:
    """Sphere function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^n x_j^2

    subject to

    .. math::
        -5.12 \\leq x_i \\leq 5.12

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -5.12 * np.ones(dim)
        self.xup = 5.12 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Sphere function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Sphere function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return np.sum(x ** 2)


class Exponential:
    """Exponential function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^n e^{jx_j} - \\sum_{j=1} e^{-5.12 j}

    subject to

    .. math::
        -5.12 \\leq x_i \\leq 5.12

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -5.12 * np.ones(dim)
        self.xup = 5.12 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Exponential function \n" +\
                             "Global optimum: f(-5.12,-5.12,...,-5.12) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Exponential function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0
        for i in range(len(x)):
            total += np.exp((i+1)*x[i-1]) - np.exp(-5.12*(i+1))
        return total


class StyblinskiTang:
    """StyblinskiTang function

    .. math::
        f(x_1,\\ldots,x_n) = \\frac{1}{2} \\sum_{j=1}^n  \
        \\left(x_j^4 -16x_j^2 +5x_j \\right)

    subject to

    .. math::
        -5 \\leq x_i \\leq 5

    Global optimum: :math:`f(-2.903534,-2.903534,...,-2.903534)=\
    -39.16599 \\cdot n`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -5 * np.ones(dim)
        self.xup = 5 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Styblinski-Tang function \n" +\
                             "Global optimum: f(-2.903534,...,-2.903534) = " +\
                             str(-39.16599*dim)
        self.min = -39.16599*dim
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the StyblinskiTang function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return 0.5*np.sum(x ** 4 - 16 * x ** 2 + 5 * x)


class Quartic:
    """Quartic function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{j=1}^n j x_j^4 + random[0,1)

    subject to

    .. math::
        -1.28 \\leq x_i \\leq 1.28

    Global optimum: :math:`f(0,0,...,0)=0+noise`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -1.28 * np.ones(dim)
        self.xup = 1.28 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Quartic function \n" +\
                             "Global optimum: f(0,0,...,0) = 0+noise"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        self.prng = random.Random()
        self.prng.seed(time())
        validate(self)

    def objfunction(self, x):
        """Evaluate the Quartic function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0.0
        for i in xrange(len(x)):
            total += (i+1.0) * x[i]**4.0
        return total + self.prng.uniform(0, 1)


class Whitley:
    """Quartic function

    .. math::
        f(x_1,\\ldots,x_n) = \\sum_{i=1}^n \\sum_{j=1}^n \
        \\left( \\frac{(100(x_i^2-x_j)^2+(1-x_j)^2)^2}{4000} \
        - \\cos(100(x_i^2-x_j)^2 + (1-x_j)^2 ) + 1 \\right)

    subject to

    .. math::
        -10.24 \\leq x_i \\leq 10.24

    Global optimum: :math:`f(1,1,...,1)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -10.24 * np.ones(dim)
        self.xup = 10.24 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional Whitley function \n" +\
                             "Global optimum: f(1,1,...,1) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the Whitley function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0
        for i in range(len(x)):
            for j in range(len(x)):
                temp = 100*((x[i]**2)-x[j]) + (1-x[j])**2
                total += (float(temp**2)/4000.0) - np.cos(temp) + 1
        return total


class SchafferF7:
    """SchafferF7 function

    .. math::
        f(x_1,\\ldots,x_n) = \\left[\\frac{1}{n-1}\\sqrt{s_i} \
        \\cdot (\\sin(50.0s_i^{\\frac{1}{5}})+1)\\right]^2

    where

    .. math::
        s_i = \\sqrt{x_i^2 + x_{i+1}^2}

    subject to

    .. math::
        -100 \\leq x_i \\leq 100

    Global optimum: :math:`f(0,0,...,0)=0`

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = -100 * np.ones(dim)
        self.xup = 100 * np.ones(dim)
        self.dim = dim
        self.info = str(dim)+"-dimensional SchafferF7 function \n" +\
                             "Global optimum: f(0,0,...,0) = 0"
        self.min = 0
        self.integer = []
        self.continuous = np.arange(0, dim)
        validate(self)

    def objfunction(self, x):
        """Evaluate the SchafferF7 function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        total = 0
        normalizer = 1.0/float(len(x)-1)
        for i in range(len(x)-1):
            si = np.sqrt(x[i]**2 + x[i+1]**2)
            total += (normalizer * np.sqrt(si) * (np.sin(50*si**0.20) + 1))**2
        return total

# ============================== Constraints ==============================


class Keane:
    """Keane's "bump" function

    .. math::
        f(x_1,\\ldots,x_n) = -\\left| \\frac{\\sum_{j=1}^n \\cos^4(x_j) - \
        2 \\prod_{j=1}^n \\cos^2(x_j)}{\\sqrt{\\sum_{j=1}^n jx_j^2}} \\right|

    subject to

    .. math::
        -100 \\leq x_i \\leq 100

    .. math::
        0.75 - \\prod_{j=1}^n x_j < 0

    .. math::
        \\sum_{j=1}^n x_j - 7.5n < 0


    Global optimum: -0.835 for large n

    :ivar dim: Number of dimensions
    :ivar xlow: Lower bound constraints
    :ivar xup: Upper bound constraints
    :ivar info: Problem information:
    :ivar min: Global optimum
    :ivar integer: Integer variables
    :ivar continuous: Continuous variables:
    :ivar constraints: Whether there are non-bound constraints
    """
    def __init__(self, dim=10):
        self.xlow = np.zeros(dim)
        self.xup = 5 * np.ones(dim)
        self.dim = dim
        self.min = -0.835
        self.integer = []
        self.continuous = np.arange(0, dim)
        self.info = str(dim)+"-dimensional Keane bump function \n" +\
                             "Global optimum: -0.835 for large n"
        validate(self)

    # Compute the value of the constraint functions at the given point
    # Returns an array of size npts x nconstraints
    def eval_ineq_constraints(self, x):
        vec = np.zeros((x.shape[0], 2))
        vec[:, 0] = 0.75 - np.prod(x)
        vec[:, 1] = np.sum(x) - 7.5 * self.dim
        return vec

    # Compute the derivative of the constraint functions
    # Returns an array of size npts x nconstraints x ndims
    def deriv_ineq_constraints(self, x):
        vec = np.zeros((x.shape[0], 2, x.shape[1]))
        for i in range(x.shape[0]):
            xx = x[i, :]
            for j in range(x.shape[1]):
                vec[i, 0, j] = -np.prod(np.hstack((xx[:j], xx[j+1:])))
                vec[i, 1, j] = 1.0
        return vec

    # Evaluate the objective function for a single data point
    def objfunction(self, x):
        """Evaluate the Keane function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
                raise ValueError('Dimension mismatch')
        n = len(x)
        return -abs((sum(np.cos(x)**4)-2 * np.prod(np.cos(x)**2)) /
                    max([1E-10, np.sqrt(np.dot(1+np.arange(n), x**2))]))


class LinearMI:
    """This is a linear mixed integer problem with non-bound constraints

    There are 5 variables, the first 3 are discrete and the last 2
    are continuous.

    Global optimum: :math:`f(1,0,0,0,0) = -1`
    """
    def __init__(self):
        self.xlow = np.zeros(5)
        self.xup = np.array([10, 10, 10, 1, 1])
        self.dim = 5
        self.min = -1
        self.integer = np.arange(0, 3)
        self.continuous = np.arange(3, 5)
        self.info = str(self.dim)+"-dimensional Linear MI \n" +\
                                  "Global optimum: f(1,0,0,0,0) = -1\n" +\
                                  str(len(self.integer)) + " integer variables"
        validate(self)

    def eval_ineq_constraints(self, x):
        vec = np.zeros((x.shape[0], 3))
        vec[:, 0] = x[:, 0] + x[:, 2] - 1.6
        vec[:, 1] = 1.333 * x[:, 1] + x[:, 3] - 3
        vec[:, 2] = - x[:, 2] - x[:, 3] + x[:, 4]
        return vec

    def objfunction(self, x):
        """Evaluate the LinearMI function  at x

        :param x: Data point
        :return: Value at x
        """
        if len(x) != self.dim:
            raise ValueError('Dimension mismatch')
        return - x[0] + 3 * x[1] + 1.5 * x[2] + 2 * x[3] - 0.5 * x[4]


if __name__ == "__main__":
    print("\n========================= Hartman3 =======================")
    fun = Hartman3()
    print(fun.info)
    print("Hartman3(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Rastrigin =======================")
    fun = Rastrigin(dim=3)
    print(fun.info)
    print("Rastrigin(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Ackley =======================")
    fun = Ackley(dim=3)
    print(fun.info)
    print("Ackley(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))

    print("\n========================= Griewank =======================")
    fun = Griewank(dim=3)
    print(fun.info)
    print("Griewank(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Rosenbrock =======================")
    fun = Rosenbrock(dim=3)
    print(fun.info)
    print("Rosenbrock(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Schwefel =======================")
    fun = Schwefel(dim=3)
    print(fun.info)
    print("Schwefel(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Sphere =======================")
    fun = Sphere(dim=3)
    print(fun.info)
    print("Sphere(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Exponential =======================")
    fun = Exponential(dim=3)
    print(fun.info)
    print("Exponential(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n======================= Styblinski-Tang =====================")
    fun = StyblinskiTang(dim=3)
    print(fun.info)
    print("StyblinskiTang(1,1,1) = " +
          str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Whitley =======================")
    fun = Whitley(dim=3)
    print(fun.info)
    print("Whitley(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Quartic =======================")
    fun = Quartic(dim=3)
    print(fun.info)
    print("Quartic(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= SchafferF7 =======================")
    fun = SchafferF7(dim=3)
    print(fun.info)
    print("SchafferF7(1,1,1) = " + str(fun.objfunction(np.array([1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Liner MI =======================")
    fun = LinearMI()
    print(fun.info)
    print("Linear_MI(1,1,1,1,1) = " +
          str(fun.objfunction(np.array([1, 1, 1, 1, 1]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))

    print("\n========================= Michalewicz =======================")
    fun = Michalewicz(dim=2)
    print(fun.info)
    print("Michalewicz(2.20, 1.57) = " +
          str(fun.objfunction(np.array([2.20, 1.57]))))
    print("Continuous variables: " + str(fun.continuous))
    print("Integer variables: " + str(fun.integer))



try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from experimental_design import *
from rbf_interpolant import *
from rbf_surfaces import *
from rs_capped import *
from search_procedure import *
from test_problems import *
from sot_sync_strategies import *
from ensemble_surrogate import EnsembleSurrogate

try:
    from GUI import GUI
except ImportError:
    pass

try:
    from mars_interpolant import MARSInterpolant
except ImportError:
    pass

try:
    from kriging_interpolant import KrigingInterpolant
except ImportError:
    pass



"""
.. module:: heurisitc_algorithms
   :synopsis: A mixed-integer GA implementation for bound constrained problems
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
"""
__author__ = 'davideriksson'

import numpy as np
from experimental_design import LatinHypercube


class GeneticAlgorithm:
    def __init__(self, function, dim, xlow, xup, intvar=[], popsize=100, ngen=100, start="LHD"):
        self.nVariables = dim
        self.nIndividuals = popsize + (popsize % 2)  # Make sure this is even
        self.lowerBoundary = np.array(xlow)
        self.upperBoundary = np.array(xup)
        self.integerVariables = np.array(intvar)
        self.start = start
        self.sigma = 0.2
        self.pMutation = 1.0/dim
        self.tournamentSize = 5
        self.pCross = 0.9
        self.numberOfGenerations = ngen
        self.function = function

    def optimize(self):
        #  Initialize population
        if isinstance(self.start, np.ndarray):
            assert self.start.shape[0] == self.nIndividuals and self.start.shape[1] == self.nVariables
            assert all(np.min(self.start, axis=0) >= self.lowerBoundary) and \
                all(np.max(self.start, axis=0) <= self.upperBoundary)
            population = self.start
        elif self.start == "LHD":
            exp_des = LatinHypercube(self.nVariables, self.nIndividuals)
            population = self.lowerBoundary + exp_des.generate_points() * \
                (self.upperBoundary - self.lowerBoundary)
        elif self.start == "Random":
            population = self.lowerBoundary + np.random.rand(self.nIndividuals, self.nVariables) *\
                (self.upperBoundary - self.lowerBoundary)
        else:
            raise AttributeError("Unknown argument for initial population")

        newpopulation = []
        #  Round positions
        if len(self.integerVariables) > 0:
            newpopulation = np.copy(population)
            population[:, self.integerVariables] = np.round(population[:, self.integerVariables])
            for i in self.integerVariables:
                ind = np.where(population[:, i] < self.lowerBoundary[i])
                population[ind, i] += 1
                ind = np.where(population[:, i] > self.upperBoundary[i])
                population[ind, i] -= 1

        #  Evaluate all individuals
        functionValues = self.function(population)
        if len(functionValues.shape) == 2:
            functionValues = np.squeeze(np.asarray(functionValues))

        # Save the best individual
        ind = np.argmin(functionValues)
        bestIndividual = np.copy(population[ind, :])
        bestValue = functionValues[ind]

        if len(self.integerVariables) > 0:
            population = newpopulation

        # Main loop
        for ngen in range(self.numberOfGenerations):
            # Do tournament selection to select the parents
            competitors = np.random.randint(0, self.nIndividuals, (self.nIndividuals, self.tournamentSize))
            ind = np.argmin(functionValues[competitors], axis=1)
            winnerIndices = np.zeros(self.nIndividuals, dtype=int)
            for i in range(self.tournamentSize):  # This loop is short
                winnerIndices[np.where(ind == i)] = competitors[np.where(ind == i), i]

            parent1 = population[winnerIndices[0:self.nIndividuals/2], :]
            parent2 = population[winnerIndices[self.nIndividuals/2:self.nIndividuals], :]

            # Averaging Crossover
            cross = np.where(np.random.rand(self.nIndividuals/2) < self.pCross)[0]
            nn = len(cross)  # Number of crossovers
            alpha = np.random.rand(nn, 1)

            # Create the new chromosomes
            parent1New = np.multiply(alpha, parent1[cross, :]) + np.multiply(1-alpha, parent2[cross, :])
            parent2New = np.multiply(alpha, parent2[cross, :]) + np.multiply(1-alpha, parent1[cross, :])
            parent1[cross, :] = parent1New
            parent2[cross, :] = parent2New
            population = np.concatenate((parent1, parent2))

            # Apply mutation
            scalefactors = self.sigma * (self.upperBoundary - self.lowerBoundary)  # Account for dimensions ranges
            perturbation = np.random.randn(self.nIndividuals, self.nVariables)  # Generate perturbations
            perturbation = np.multiply(perturbation, scalefactors)  # Scale accordingly
            perturbation = np.multiply(perturbation, (np.random.rand(self.nIndividuals,  # Determine where to perturb
                                                                     self.nVariables) < self.pMutation))
            population += perturbation  # Add perturbation
            population = np.maximum(np.reshape(self.lowerBoundary, (1, self.nVariables)), population)
            population = np.minimum(np.reshape(self.upperBoundary, (1, self.nVariables)), population)

            # Round chromosomes
            newpopulation = []
            #  Round positions
            if len(self.integerVariables) > 0:
                newpopulation = np.copy(population)
                population[:, self.integerVariables] = np.round(population[:, self.integerVariables])
                for i in self.integerVariables:
                    ind = np.where(population[:, i] < self.lowerBoundary[i])
                    population[ind, i] += 1
                    ind = np.where(population[:, i] > self.upperBoundary[i])
                    population[ind, i] -= 1

            # Keep the best individual
            population[0, :] = bestIndividual

            #  Evaluate all individuals
            functionValues = self.function(population)
            if len(functionValues.shape) == 2:
                functionValues = np.squeeze(np.asarray(functionValues))

            # Save the best individual
            ind = np.argmin(functionValues)
            bestIndividual = np.copy(population[ind, :])
            bestValue = functionValues[ind]

            if len(self.integerVariables) > 0:
                population = newpopulation

        return bestIndividual, bestValue

def main():

    dim = 30

    # Vectorized Ackley function in dim dimensions
    def objfunction(x):
        return -20.0*np.exp(-0.2*np.sqrt(np.sum(x**2, axis=1)/dim)) - \
            np.exp(np.sum(np.cos(2.0*np.pi*x), axis=1)/dim)

    ga = GeneticAlgorithm(objfunction, dim, -15*np.ones(dim), 20*np.ones(dim), popsize=100, ngen=100, start="SLHD")
    xBest, fBest = ga.optimize()

    # Print the best solution found
    print("\nBest function value: %f" % fBest)
    print("Best solution: " % xBest)
    np.set_printoptions(suppress=True)
    print(xBest)

if __name__ == "__main__":
    main()




"""
.. module:: search_procedure
  :synopsis: ways of finding the next point to evaluate
.. moduleauthor:: David Eriksson <dme65@cornell.edu>
                 David Bindel <bindel@cornell.edu>

:Module: search_procedure
:Author: David Eriksson <dme65@cornell.edu>,
    David Bindel <bindel@cornell.edu>

We currently only support the weighted distance merit function.

We support the following methods for minimizing on the surface:

Candidate based methods:
    - CandidateSRBF: Generate candidate points around the best point
    - CandidateDYCORS: Uses a DDS strategy with a cap on the lowest probability
    - CandidateUniform: Sample the candidate points uniformly in the domain
    - CandidateSRBF_INT: Uses CandidateSRBF but only
        perturbs the integer variables
    - CandidateSRBF_CONT: Uses CandidateSRBF but only
        perturbs the continuous variables
    - CandidateDYCORS_INT: Uses CandidateSRBF but only
        perturbs the integer variables
    - CandidateDYCORS_CONT: Uses CandidateSRBF but only
        perturbs the continuous variables
    - CandidateUniform_CONT: Sample the continuous variables of
        the candidate points uniformly in the domain
    - CandidateUniform_INT: Sample the integer variables of the
        candidate points uniformly in the domain

We also support using multiple of these strategies and cycle
    between them which we call MultiSearchStrategy
    - MultiSearchStrategy
"""

import math
import numpy as np
import scipy.spatial as scp
from heuristic_algorithms import GeneticAlgorithm as GA
from scipy.optimize import minimize
import scipy.stats as stats

# ========================= Useful helpers =======================


def to_unit_box(x, data):
    return (np.copy(x) - data.xlow) / (data.xup - data.xlow)


def from_unit_box(x, data):
        return data.xlow + (data.xup - data.xlow) * np.copy(x)


def unit_rescale(xx):
    """Shift and rescale elements of a vector to the unit interval
    """
    xmax = np.amax(xx)
    xmin = np.amin(xx)
    if xmax == xmin:
        return np.ones(xx.shape)
    else:
        return (xx-xmin)/(xmax-xmin)


def round_vars(data, x):
    """Round integer variables to closest integer
    """
    if len(data.integer) > 0:
        # Map to original range
        x = from_unit_box(x, data)
        # Round the original ranged integer variables
        x[:, data.integer] = np.round(x[:, data.integer])
        # Make sure we don't violate the bound constraints
        for i in data.integer:
            ind = np.where(x[:, i] < data.xlow[i])
            x[ind, i] += 1
            ind = np.where(x[:, i] > data.xup[i])
            x[ind, i] -= 1
        # Map back to [0, 1] and return
        return to_unit_box(x, data)
    return x

# ========================= MultiSearch =======================


class MultiSearchStrategy(object):
    """ A collection of Search Strategies and weights so that the user
        can use multiple search strategies for the same optimization
        problem. This object keeps an internal list of proposed points
        in order to be able to compute the minimum distance from a point
        to all proposed evaluations. This list has to be reset each time
        the optimization algorithm restarts.

        :ivar search_strategies: List of search strategies
        :ivar weights: List of integer weights that will be iterated in
            order to determine what search strategy to use next.
            If for example the weights [0 0 1] are used, then
            search strategy 0 will be used two times, then
            strategy 1 is used after which the cycling restarts.
        :ivar proposed_points: List of all points proposed by this strategy
        :ivar currentWeight: Current weights
        :ivar data: Optimization problem object
        :ivar avoid: Points to avoid
    """
    def __init__(self, strategy_list, weight_strategy):
        # Check so that the weights are correct
        if weight_strategy is None:
            weight_strategy = range(len(strategy_list))
        if (not all(isinstance(ii, int) for ii in weight_strategy)) or \
                np.min(weight_strategy) < 0 or \
                np.max(weight_strategy) > len(strategy_list)-1:
            raise ValueError("Incorrect weights!!")
        self.search_strategies = strategy_list
        self.weights = weight_strategy
        self.currentWeight = 0
        self.proposed_points = None
        self.data = strategy_list[0].data
        self.avoid = None
        self.budget = None
        self.n0 = None
        self.issync = None
        self.fhat = None

    def init(self, start_sample, budget, issync, fhat, avoid=None):
        """Initialize the multi-search strategy

        :param start_sample: Points in the initial design that
            will be evaluated before the adaptive sampling starts
        :param avoid: Points to avoid
        """
        self.proposed_points = start_sample
        self.avoid = avoid
        self.n0 = start_sample.shape[0]
        self.budget = budget
        self.fhat = fhat
        self.issync = issync
        for i in range(len(self.search_strategies)):
            self.search_strategies[i].init(self.proposed_points, budget, issync, fhat, avoid)

    def remove_point(self, x):
        """Remove x from proposed_points. Useful if x was never evaluated.

        :param x: Point to remove
        """
        idx = np.sum(np.abs(self.proposed_points - x), axis=1).argmin()
        self.proposed_points = np.delete(self.proposed_points, idx, axis=0)
        for i in range(len(self.search_strategies)):
            self.search_strategies[i].proposed_points = self.proposed_points

    def make_points(self, xbest, sigma, evals, derivs):
        """Create new candidate points. This call is ignored by the optimization
        based search strategies.


        :param xbest: Best solution found
        :param sigma: Current stdDev used to generate candidate points
        :param evals: Routine for predicting function values
        :param deriv: Routine for predicting derivatives
        """
        if self.issync:
            for i in range(len(self.search_strategies)):
                self.search_strategies[i].make_points(xbest, sigma, evals, derivs)
        else:
            weight = self.weights[self.currentWeight]
            self.search_strategies[weight].make_points(xbest, sigma, evals, derivs)

    def next(self):
        """Generate the next proposed point from the current search strategy,
        update the list of proposed points and move the counter forward to
        the next search strategy

        :return: Next point to evaluate
        """
        weight = self.weights[self.currentWeight]
        xnew = np.reshape(self.search_strategies[weight].next(),
                          (-1, self.data.dim))
        self.proposed_points = self.search_strategies[weight].proposed_points
        # Update samples
        for i in range(len(self.search_strategies)):
            self.search_strategies[i].proposed_points = self.proposed_points
            # Append xsample to all other candidate based methods
            if self.search_strategies[i].usecand and i != weight:
                if self.search_strategies[i].xsample is None:
                    self.search_strategies[i].xsample = []
                    self.search_strategies[i].xsample.append(xnew)
                else:
                    self.search_strategies[i].xsample.append(xnew)

        self.currentWeight = (self.currentWeight + 1) % len(self.weights)

        return xnew


# ====================== Candidate based search methods =====================

def candidate_merit_weighted_distance(cand):
    """Weighted distance merit functions for the candidate points based methods
    """

    ii = cand.nextWeight
    if cand.xsample:
        ds = scp.distance.cdist(cand.xcand, np.atleast_2d(cand.xsample[-1]))
        cand.dmerit = np.minimum(cand.dmerit, ds)
    if cand.avoid:
        for _, xavoid in cand.avoid.iteritems():
            xavoid = np.reshape(xavoid, (1, xavoid.shape[0]))
            davoid = scp.distance.cdist(cand.xcand, xavoid)
            cand.dmerit = np.minimum(cand.dmerit, davoid)

    weight = cand.weights[(ii+len(cand.weights)) % len(cand.weights)]
    merit = weight*cand.fhvals + \
        (1-weight)*(1.0 - unit_rescale(cand.dmerit))
    merit[cand.dmerit < cand.dtol] = np.inf
    jj = np.argmin(merit)
    cand.fhvals[jj] = np.inf
    cand.xsample.append(cand.xcand[jj, :])
    cand.nextWeight += 1
    return cand.xcand[jj, :]


class CandidateSRBF(object):
    """This is an implementation of the candidate points method that is
    proposed in the first SRBF paper. Candidate points are generated
    by making normally distributed perturbations with stdDev sigma
    around the best solution

    :ivar usecand: Indicates that this method is candidate based
    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar xsample: The proposed evaluations since
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """

    usecand = True

    def __init__(self, data, numcand=None):
        """Initialize the multi-search strategy

        :param data: Optimization object
        :param numcand: Number of candidate points to generate
        """
        self.data = data
        self.fhat = None
        self.avoid = None
        self.xrange = np.ones(data.xlow.shape)
        self.dtol = 1e-3 * math.sqrt(data.dim)
        self.weights = np.array([0.3, 0.5, 0.8, 0.95])
        self.dmerit = None
        self.fhvals = None
        self.xsample = None
        self.xcand = None
        self.proposed_points = None
        self.nextWeight = 0
        self.numcand = numcand
        if self.numcand is None:
            self.numcand = min([5000, 100*data.dim])
        self.budget = None
        self.n0 = None
        self.issync = None

    def init(self, start_sample, budget, issync, fhat, avoid=None):
        self.proposed_points = start_sample
        self.avoid = avoid
        self.n0 = start_sample.shape[0]
        self.budget = budget
        self.fhat = fhat
        self.issync = issync

    def remove_point(self, x):
        """Remove x from the list of proposed points.
        Useful if x was never evaluated.

        :param x: Point to be removed
        """
        idx = np.sum(np.abs(self.proposed_points - x), axis=1).argmin()
        self.proposed_points = np.delete(self.proposed_points, idx, axis=0)

    def next(self):
        """Propose a new point to evaluate and update list of proposed points

        :return: Next point to evaluate
        """
        xnew = candidate_merit_weighted_distance(self)  # FIXME, hard-coded
        self.proposed_points = np.vstack((self.proposed_points,
                                          np.asarray(xnew)))
        return xnew

    def _generate_cand(self, scalefactors, xbest, subset):
        xcand = np.ones((self.numcand,  self.data.dim)) * xbest
        for i in subset:
            lower, upper = 0.0, 1.0
            ssigma = scalefactors[i]
            xcand[:, i] = stats.truncnorm.rvs(
                (lower - xbest[i]) / ssigma, (upper - xbest[i]) / ssigma,
                loc=xbest[i], scale=ssigma, size=self.numcand)
        self.xcand = round_vars(self.data, xcand)

    def make_points(self, xbest, sigma, evals, derivs=None, subset=None):
        """Create new candidate points based on the best
        solution and the current value of sigma.

        :param xbest: Best solution found so far
        :param sigma: Current radius, i.e. stdDev used
            to generate candidate points
        :param maxeval: Ignored by this method
        :param issync: Ignored by this method
        :param subset: Dimensions that will be perturbed
        """

        if subset is None:
            subset = np.arange(0, self.data.dim)
        scalefactors = sigma * self.xrange

        # Make sure that the scale factors are correct for
        # the integer variables (at least 1)
        ind = np.intersect1d(self.data.integer, subset)
        if len(ind) > 0:
            scalefactors[ind] = np.maximum(scalefactors[ind], 1.0 / (self.data.xup[ind] - self.data.xlow[ind]))

        # Generate candidate points
        self._generate_cand(scalefactors, xbest, subset)

        devals = scp.distance.cdist(self.xcand, self.proposed_points)

        # Since the response surface may have the centers in a  different order than the
        # evaluations were proposed we need to map the distances to the centers correctly
        mapping = []
        for i in range(self.fhat.nump):
            mapping.append(np.sum(np.abs(self.proposed_points - self.fhat.x[i, :]), axis=1).argmin())

        fhvals = evals(self.xcand, devals[:, mapping].T, scaling=True)

        self.dmerit = np.amin(np.asmatrix(devals), axis=1)
        self.fhvals = unit_rescale(fhvals)
        self.xsample = []


class CandidateUniform(CandidateSRBF):
    """Create Candidate points by sampling uniformly in the domain

    :ivar usecand: Indicates that this method is candidate based
    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar xsample: The proposed evaluations since
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """

    def _generate_cand(self, scalefactors, xbest, subset):
        xcand = np.ones((self.numcand, self.data.dim)) * xbest
        xcand[:, subset] = np.random.uniform(
            0.0, 1.0, (self.numcand, len(subset)))

        self.xcand = round_vars(self.data, xcand)


class CandidateDYCORS(CandidateSRBF):
    """This is an implementation of DyCORS method to generate
    candidate points. The DyCORS method uses the DDS algorithm
    which only perturbs a subset of the dimensions when
    perturbing the best solution. The probability for a
    dimension to be perturbed decreases after each
    evaluation and is capped in order to guarantee
    global convergence.

    :ivar usecand: Indicates that this method is candidate based
    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar xsample: The proposed evaluations since
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """
    def __init__(self, data, numcand=None):
        """Initialize the DYCORS strategy

        :param data: Optimization object
        :param numcand:  Number of candidate points to generate
        """
        CandidateSRBF.__init__(self, data, numcand=numcand)
        self.minprob = np.min([1.0, 1.0/self.data.dim])

        def probfun(numevals, budget):
            return min([20.0/data.dim, 1.0]) * (1.0-(np.log(numevals + 1.0) / np.log(budget)))
        self.probfun = probfun

    def _generate_cand(self, scalefactors, xbest, subset):
        ddsprob = self.probfun(self.proposed_points.shape[0] - self.n0, self.budget - self.n0)
        ddsprob = np.max([self.minprob, ddsprob])

        nlen = len(subset)
        ar = (np.random.rand(self.numcand, nlen) < ddsprob)
        ind = np.where(np.sum(ar, axis=1) == 0)[0]
        ar[ind, np.random.randint(0, nlen - 1, size=len(ind))] = 1

        xcand = np.ones((self.numcand, self.data.dim)) * xbest
        for i in range(nlen):
            lower, upper = 0.0, 1.0
            ssigma = scalefactors[subset[i]]
            ind = np.where(ar[:, i] == 1)[0]
            xcand[ind, subset[i]] = stats.truncnorm.rvs(
                (lower - xbest[subset[i]]) / ssigma, (upper - xbest[subset[i]]) / ssigma,
                loc=xbest[subset[i]], scale=ssigma, size=len(ind))

        self.xcand = round_vars(self.data, xcand)


class CandidateDDS(CandidateDYCORS):
    def __init__(self, data, numcand=None):
        CandidateDYCORS.__init__(self, data, numcand=numcand)
        self.weights = np.array([1.0])
        self.numcand = max([0.5*data.dim, 2])

        def probfun(numevals, budget):
            return 1.0-(np.log(numevals + 1.0) / np.log(budget))
        self.probfun = probfun

    def make_points(self, xbest, sigma, evals, derivs=None, subset=None):
        CandidateDYCORS.make_points(self, xbest, 0.2, evals, derivs, subset)


class CandidateSRBF_INT(CandidateSRBF):
    """Candidate points are generated by perturbing
    ONLY the discrete variables using the SRBF strategy

    :ivar usecand: Indicates that this method is candidate based
    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar xsample: The proposed evaluations since
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """
    def make_points(self, xbest, sigma, evals, derivs=None, subset=None):
        if len(self.data.integer) > 0:
            CandidateSRBF.make_points(self, xbest, sigma,
                                      evals, derivs, subset=self.data.integer)
        else:
            CandidateSRBF.make_points(self, xbest, sigma,
                                      evals, derivs, subset=self.data.continuous)


class CandidateDYCORS_INT(CandidateDYCORS):
    """Candidate points are generated by perturbing ONLY the discrete
    variables using the DyCORS strategy

    :ivar usecand: Indicates that this method is candidate based
    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar xsample: The proposed evaluations since
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """
    def make_points(self, xbest, sigma, evals, derivs=None, subset=None):
        if len(self.data.integer) > 0:
            CandidateDYCORS.make_points(self, xbest, sigma,
                                        evals, derivs=None, subset=self.data.integer)
        else:
            CandidateDYCORS.make_points(self, xbest, sigma,
                                        evals, derivs=None, subset=self.data.continuous)


class CandidateUniform_INT(CandidateUniform):
    """Candidate points are generated by perturbing ONLY
    the discrete variables using the uniform perturbations

    :ivar usecand: Indicates that this method is candidate based
    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar xsample: The proposed evaluations since
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """
    def make_points(self, xbest, sigma, evals, derivs=None, subset=None):
        if len(self.data.integer) > 0:
            CandidateUniform.make_points(self, xbest, sigma, evals,
                                         derivs=None, subset=self.data.integer)
        else:
            CandidateUniform.make_points(self, xbest, sigma, evals,
                                         derivs=None, subset=self.data.continuous)


class CandidateSRBF_CONT(CandidateSRBF):
    """Candidate points are generated by perturbing ONLY
    the continuous variables using the SRBF strategy

    :ivar usecand: Indicates that this method is candidate based
    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar xsample: The proposed evaluations since
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """
    def make_points(self, xbest, sigma, evals, derivs=None, subset=None):
        if len(self.data.continuous) > 0:
            CandidateSRBF.make_points(self, xbest, sigma, evals,
                                      derivs=None, subset=self.data.continuous)
        else:
            CandidateSRBF.make_points(self, xbest, sigma, evals,
                                      derivs=None, subset=self.data.integer)


class CandidateDYCORS_CONT(CandidateDYCORS):
    """Candidate points are generated by perturbing ONLY
    the continuous variables using the DyCORS strategy

    :ivar usecand: Indicates that this method is candidate based
    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar xsample: The proposed evaluations since
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """
    def make_points(self, xbest, sigma, evals, derivs=None, subset=None):
        if len(self.data.continuous) > 0:
            CandidateDYCORS.make_points(self, xbest, sigma, evals,
                                        derivs=None, subset=self.data.continuous)
        else:
            CandidateDYCORS.make_points(self, xbest, sigma, evals,
                                        derivs=None, subset=self.data.integer)


class CandidateUniform_CONT(CandidateUniform):
    """Candidate points are generated by perturbing ONLY
    the continuous variables using uniform perturbations

    :ivar usecand: Indicates that this method is candidate based
    :ivar data: Optimization object
    :ivar weights: Weights used in the merit function
    :ivar numcand: Number of candidate points to generate
    :ivar xsample: The proposed evaluations since
    :ivar proposed_points: List of points proposed by any search strategy
        since the last restart
    """

    def make_points(self, xbest, sigma, evals, derivs=None, subset=None):
        if len(self.data.continuous) > 0:
            CandidateUniform.make_points(self, xbest, sigma, evals,
                                         derivs=None, subset=self.data.continuous)
        else:
            CandidateUniform.make_points(self, xbest, sigma, evals,
                                         derivs=None, subset=self.data.integer)

################################## Optimization based strategies ###################################

class GeneticAlgorithm(object):

    usecand = False

    def __init__(self, data):
        """Initialize the multi-search strategy

        :param data: Optimization object
        :param numcand: Number of candidate points to generate
        """
        self.data = data
        self.fhat = None
        self.avoid = None
        self.dtol = 1e-3 * math.sqrt(data.dim)
        self.dmerit = None
        self.proposed_points = None
        self.evals = None
        self.derivs = None
        self.budget = None
        self.numevals = None
        self.issync = None

    def init(self, start_sample, budget, issync, fhat, avoid=None):
        self.proposed_points = start_sample
        self.avoid = avoid
        self.budget = budget
        self.numevals = 0
        self.fhat = fhat
        self.issync = issync

    def remove_point(self, x):
        idx = np.sum(np.abs(self.proposed_points - x), axis=1).argmin()
        self.proposed_points = np.delete(self.proposed_points, idx, axis=0)

    def next(self, objfunction=None):

        # Find a new point
        if objfunction is None:
            def objfunction(x):
                return self.evals(to_unit_box(x, self.data), scaling=False)

        ga = GA(objfunction, self.data.dim, self.data.xlow,
                self.data.xup, popsize=max([2*self.data.dim, 100]), ngen=100)
        xnew, fbest = ga.optimize()
        xnew = to_unit_box(xnew, self.data)

        dist = np.atleast_2d(np.min(
            scp.distance.cdist(self.proposed_points, np.atleast_2d(xnew)), axis=1)).T

        if np.min(dist) < self.dtol:
            # Perturb the best solution until we satisfy the tolerance
            d = 0.0
            xbest = None
            while d < self.dtol:
                xbest = xnew + self.dtol * np.random.randn(1, self.data.dim)
                xbest = np.maximum(xbest, np.zeros(self.data.dim))
                xbest = np.minimum(xbest, np.ones(self.data.dim))
                d = np.atleast_2d(np.min(
                    scp.distance.cdist(self.proposed_points, np.atleast_2d(xbest)), axis=1)).T.min()
            xnew = xbest

        self.proposed_points = np.vstack((self.proposed_points, np.asarray(xnew)))
        return xnew

    def make_points(self, xbest, sigma, evals, derivs=None):
        self.evals = evals
        self.derivs = derivs


class MultiStartGradient(object):

    usecand = False

    """ A wrapper around the scipy.optimize implementations of box-constrained
        gradient based minimization.

    Attributes:
        usecand: Indicates that this method is NOT candidate based
        data: Optimization object
        fhat: Original response surface object.
        proposed_points: List of points proposed by any search strategy
                         since the last restart
        objfun: The merit function to minimize (needs to have a gradient available)
        bounds: n x 2 matrix with lower and upper bound constraints
        numrestarts: Number of random starting points
        method: What optimization method to use. The following options are available:
            - L-BFGS-B: Quasi-Newton method of
                        Broyden, Fletcher, Goldfarb, and Shanno (BFGS)
            - TNC:      Truncated Newton algorithm

        Note: SLSQP is supposed to work with bound constraints but for some reason it
              sometimes violates the constraints anyway.
    """
    def __init__(self, data, method='L-BFGS-B', numrestarts=30):
        """Initialize the Multi-Start Gradient object

        Args:
            data: Optimization object
            method: What optimization method to use (see above)
            numrestarts: Number of random starting points
        """
        self.data = data
        self.fhat = None
        self.avoid = None
        self.dtol = 1e-3 * math.sqrt(data.dim)
        self.proposed_points = None
        self.evals = None
        self.derivs = None
        self.budget = None
        self.numevals = None
        self.issync = None
        self.bounds = np.zeros((self.data.dim, 2))
        self.bounds[:, 0] = self.data.xlow
        self.bounds[:, 1] = self.data.xup
        self.numrestarts = numrestarts
        self.xbest = None
        if (method == 'TNC') or (method == 'L-BFGS-B'):
            self.method = method
        else:
            self.method = 'L-BFGS-B'

    def init(self, start_sample, budget, issync, fhat, avoid=None):
        self.proposed_points = start_sample
        self.avoid = avoid
        self.budget = budget
        self.numevals = 0
        self.fhat = fhat
        self.issync = issync

    def remove_point(self, x):
        """ Remove x from self.porposed_points."""
        idx = np.sum(np.abs(self.proposed_points - x), axis=1).argmin()
        self.proposed_points = np.delete(self.proposed_points, idx, axis=0)

    def make_points(self, xbest, sigma, evals, derivs):
        """The method doesn't use candidate points so this method is not used """
        self.evals = evals
        self.derivs = derivs
        self.xbest = xbest

    def next(self):
        """Propose a new point to evaluate and update list of proposed points """
        def eval(x):
            return self.evals(np.atleast_2d(to_unit_box(x, self.data)))[0, 0]

        def deriv(x):
            return self.derivs(to_unit_box(x, self.data)).ravel()

        self.fvals = np.zeros(self.numrestarts)
        self.xvals = np.zeros((self.numrestarts, self.data.dim))
        self.dists = np.zeros(self.numrestarts)

        for i in range(self.numrestarts):
            if i == 0:
                x0 = np.array(from_unit_box(self.xbest, self.data))
            else:
                x0 = np.random.uniform(self.data.xlow, self.data.xup)

            res = minimize(eval, x0, method=self.method,
                           jac=deriv, bounds=self.bounds)

            # Compute the distance to the proposed points
            xx = np.atleast_2d(res.x)
            xx = to_unit_box(xx, self.data)
            dist = np.atleast_2d(np.min(
                scp.distance.cdist(self.proposed_points, xx), axis=1)).T

            self.fvals[i] = res.fun
            self.xvals[i, :] = xx
            self.dists[i] = dist.min()
        if self.dists.max() > self.dtol:
            xbest = None
            fbest = np.inf
            for i in range(self.numrestarts):
                if self.dists[i] > self.dtol and self.fvals[i] < fbest:
                    xbest = self.xvals[i, :]
                    fbest = self.fvals[i]
        else:
            # Perturb the best point
            d = 0.0
            while d < self.dtol:
                xbest = self.xvals[np.argmin(self.fvals), :] + self.dtol * np.random.randn(1, self.data.dim)
                xbest = np.maximum(xbest, np.zeros(self.data.dim))
                xbest = np.minimum(xbest, np.ones(self.data.dim))
                d = np.atleast_2d(np.min(
                    scp.distance.cdist(self.proposed_points, np.atleast_2d(xbest)), axis=1)).T.min()

        self.proposed_points = np.vstack((self.proposed_points, np.asarray(xbest)))
        return xbest.ravel()

if __name__ == "__main__":
    # Test DYCORS
    dim = 10
    maxeval = 100
    initeval = 80
    from test_problems import Ackley
    from experimental_design import LatinHypercube
    from rbf_interpolant import RBFInterpolant
    from rbf_surfaces import CubicRBFSurface
    xbest = np.ones(dim)
    data = Ackley(dim)
    cand = CandidateDYCORS(data, 100*dim)
    exp_des = LatinHypercube(dim, npts=initeval)
    initpoints = exp_des.generate_points()
    rbf = RBFInterpolant(surftype=CubicRBFSurface)
    for i in range(initeval):
        rbf.add_point(initpoints[i, :], data.objfunction(initpoints[i, :]))
    cand.init(initpoints, maxeval, True, None)

    def evals(x, scaling):
        return rbf.evals(x)

    cand.make_points(xbest, 0.02, evals, maxeval)

    import matplotlib.pyplot as plt
    plt.plot(cand.xcand[:, 0], cand.xcand[:, 1], 'ro')
    plt.ylabel('Dimension 1')
    plt.xlabel('Dimension 2')
    plt.show()



"""
.. module:: sot_sync_strategies
   :synopsis: Parallel synchronous optimization strategy
.. moduleauthor:: David Bindel <bindel@cornell.edu>,
    David Eriksson <dme65@cornell.edu>

:Module: sot_sync_strategies
:Author: David Bindel <bindel@cornell.edu>,
    David Eriksson <dme65@cornell.edu>

Synchronous strategies for Stochastic RBF
"""

from __future__ import print_function
import numpy as np
import math
import logging
from experimental_design import SymmetricLatinHypercube, LatinHypercube
from search_procedure import round_vars, CandidateDYCORS, \
    MultiSearchStrategy, GeneticAlgorithm
from poap.strategy import BaseStrategy, RetryStrategy
from rbf_surfaces import CubicRBFSurface
from rbf_interpolant import RBFInterpolant


# Get module-level logger
logger = logging.getLogger(__name__)


class SyncStrategyNoConstraints(BaseStrategy):
    """Parallel synchronous optimization strategy without non-bound constraints.

    This class implements the parallel synchronous SRBF strategy
    described by Regis and Shoemaker.  After the initial experimental
    design (which is embarrassingly parallel), the optimization
    proceeds in phases.  During each phase, we allow nsamples
    simultaneous function evaluations.  We insist that these
    evaluations run to completion -- if one fails for whatever reason,
    we will resubmit it.  Samples are drawn randomly from around the
    current best point, and are sorted according to a merit function
    based on distance to other sample points and predicted function
    values according to the response surface.  After several
    successive significant improvements, we increase the sampling
    radius; after several failures to improve the function value, we
    decrease the sampling radius.  We restart once the sampling radius
    decreases below a threshold.
    """

    def __init__(self, worker_id, data, response_surface, maxeval, nsamples,
                 exp_design=None, search_procedure=None, extra=None):
        """Initialize the optimization strategy.

        :param worker_id: Start ID in a multistart setting
        :param data: Problem parameter data structure
        :param response_surface: Surrogate model object
        :param maxeval: Function evaluation budget
        :param nsamples: Number of simultaneous fevals allowed
        :param exp_design: Experimental design
        :param search_procedure: Search procedure for finding
            points to evaluate
        :param extra: Points to be added to the experimental design
        """

        if self.__class__.__name__ == "SyncStrategyNoConstraints":
            assert not hasattr(data, "eval_ineq_constraints"), "Objective function has constraints,\n" \
                "SyncStrategyNoConstraints can't handle constraints"

        self.worker_id = worker_id
        self.data = data
        self.fhat = response_surface
        if self.fhat is None:
            self.fhat = RBFInterpolant(surftype=CubicRBFSurface, eta=1e-8, maxp=maxeval)
        self.maxeval = maxeval
        self.nsamples = nsamples
        self.extra = extra

        # Default to generate sampling points using Symmetric Latin Hypercube
        self.design = exp_design
        if self.design is None:
            if self.data.dim > 50:
                self.design = LatinHypercube(data.dim, data.dim+1)
            else:
                self.design = SymmetricLatinHypercube(data.dim, 2*(data.dim+1))

        self.xrange = np.asarray(data.xup - data.xlow)

        # algorithm parameters
        self.sigma_min = 0.005
        self.sigma_max = 0.2
        self.sigma_init = 0.2

        self.failtol = max(5, data.dim)
        self.succtol = 3

        self.numeval = 0
        self.status = 0
        self.sigma = 0
        self.resubmitter = RetryStrategy()
        self.xbest = None
        self.fbest = np.inf
        self.fbest_old = None

        # Set up search procedures and initialize
        self.search = search_procedure
        if self.search is None:
            self.search = CandidateDYCORS(data)

        # Start with first experimental design
        self.sample_initial()

    def to_unit_box(self, x):
        return (np.copy(x) - self.data.xlow) / (self.data.xup - self.data.xlow)

    def from_unit_box(self, x):
        return self.data.xlow + (self.data.xup - self.data.xlow) * np.copy(x)

    def log_completion(self, record):
        """Record a completed evaluation to the log.

        :param record: Record of the function evaluation
        """
        xstr = np.array_str(record.params[0], max_line_width=np.inf,
                            precision=5, suppress_small=True)
        logger.info("Feasible {:.3e} @ {}".format(record.value, xstr))

    def adjust_step(self):
        """Adjust the sampling radius sigma.

        After succtol successful steps, we cut the sampling radius;
        after failtol failed steps, we double the sampling radius.

        :ivar Fnew: Best function value in new step
        :ivar fbest: Previous best function evaluation
        """
        # Initialize if this is the first adaptive step
        if self.fbest_old is None:
            self.fbest_old = self.fbest
            return

        # Check if we succeeded at significant improvement
        if self.fbest < self.fbest_old - 1e-3 * math.fabs(self.fbest_old):
            self.status = max(1, self.status + 1)
        else:
            self.status = min(-1, self.status - 1)
        self.fbest_old = self.fbest

        # Check if step needs adjusting
        if self.status <= -self.failtol:
            self.status = 0
            self.sigma /= 2
            logger.info("Reducing sigma")
        if self.status >= self.succtol:
            self.status = 0
            self.sigma = min([2.0 * self.sigma, self.sigma_max])
            logger.info("Increasing sigma")

    def sample_initial(self):
        """Generate and queue an initial experimental design.
        """
        if self.numeval == 0:
            logger.info("=== Start ===")
        else:
            logger.info("=== Restart ===")
        self.fhat.reset()
        self.sigma = self.sigma_init
        self.status = 0
        self.xbest = None
        self.fbest_old = None
        self.fbest = np.inf
        self.fhat.reset()
        start_sample = self.design.generate_points()
        # Add extra evaluation points provided by the user
        if self.extra is not None:
            start_sample = np.vstack((start_sample, self.to_unit_box(self.extra)))

        start_sample = round_vars(self.data, start_sample)
        for j in range(min(start_sample.shape[0], self.maxeval - self.numeval)):
            proposal = self.propose_eval(self.from_unit_box(start_sample[j, :]))
            self.resubmitter.rput(proposal)

        self.search.init(start_sample, self.maxeval - self.numeval, True, self.fhat)

    def evals(self, xx, d=None, scaling=False):
        """Predict function values

        As a measure of promising function values we let all infeasible points
        have the value of the feasible candidate point with the worst function
        value, since large penalties makes it impossible to distinguish
        between feasible points.

        :param xx: Data points
        :return: Predicted function values
        """
        return self.fhat.evals(xx, d)

    def derivs(self, xx, d=None):
        return np.atleast_2d(self.fhat.deriv(xx, d))

    def sample_adapt(self):
        """Generate and queue samples from the search strategy
        """
        self.adjust_step()
        nsamples = min(self.nsamples, self.maxeval - self.numeval)
        self.search.make_points(self.xbest, self.sigma, self.evals, self.derivs)
        for _ in range(nsamples):
            proposal = self.propose_eval(np.ravel(self.from_unit_box(self.search.next())))
            self.resubmitter.rput(proposal)

    def start_batch(self):
        """Generate and queue a new batch of points
        """
        if self.sigma < self.sigma_min:
            self.sample_initial()
        else:
            self.sample_adapt()

    def propose_action(self):
        """Propose an action
        """
        if self.numeval == self.maxeval:
            return self.propose_terminate()
        elif self.resubmitter.num_eval_outstanding == 0:
            self.start_batch()
        return self.resubmitter.get()

    def on_complete(self, record):
        """Handle completed function evaluation.

        When a function evaluation is completed we need to ask the constraint
        handler if the function value should be modified which is the case for
        say a penalty method. We also need to print the information to the
        logfile, update the best value found so far and notify the GUI that
        an evaluation has completed.

        :param record: Evaluation record
        """

        self.log_completion(record)
        self.numeval += 1
        record.worker_id = self.worker_id
        record.worker_numeval = self.numeval
        self.fhat.add_point(self.to_unit_box(record.params[0]), record.value)
        if record.value < self.fbest:
            self.xbest = self.to_unit_box(record.params[0])
            self.fbest = record.value


class SyncStrategyPenalty(SyncStrategyNoConstraints):
    """Parallel synchronous optimization strategy with non-bound constraints.

    This is an extension of SyncStrategyNoConstraints that also works with
    bound constraints. We currently only allow inequality constraints, since
    the candidate based methods don't work well with equality constraints.
    We also assume that the constraints are cheap to evaluate, i.e., so that
    it is easy to check if a given point is feasible. More strategies that
    can handle expensive constraints will be added.

    We use a penalty method in the sense that we try to minimize:

    .. math::
        f(x) + \\mu \\sum_j (\\max(0, g_j(x))^2

    where :math:`g_j(x) \\leq 0` are cheap inequality constraints. As a
    measure of promising function values we let all infeasible points have
    the value of the feasible candidate point with the worst function value,
    since large penalties makes it impossible to distinguish between feasible
    points.

    When it comes to the value of :math:`\\mu`, just choose a very large value.


    """

    def __init__(self, worker_id, data, response_surface, maxeval, nsamples,
                 exp_design=None, search_procedure=None, extra=None,
                 penalty=1e6):
        """Initialize the optimization strategy.

        :param worker_id: Start ID in a multistart setting
        :param data: Problem parameter data structure
        :param response_surface: Surrogate model object
        :param maxeval: Function evaluation budget
        :param nsamples: Number of simultaneous fevals allowed
        :param exp_design: Experimental design
        :param search_procedure: Search procedure for finding
            points to evaluate
        :param extra: Points to be added to the experimental design
        :param penalty: Penalty for violating constraints
        """
        SyncStrategyNoConstraints.__init__(self,  worker_id, data,
                                           response_surface, maxeval,
                                           nsamples, exp_design,
                                           search_procedure, extra)

        if self.__class__.__name__ == "SyncStrategyPenalty":
            assert hasattr(data, "eval_ineq_constraints"), "Objective function has no constraints"

        self.penalty = penalty

    def penalty_fun(self, xx):
        """Computes the penalty for constraints violation

        :param xx: Points to compute the penalty for
        :return: Penalty for constraint violations
        """
        # Get the constraint violations
        vec = np.array(self.data.eval_ineq_constraints(self.from_unit_box(xx)))
        # Now apply the penalty for the constraint violation
        vec[np.where(vec < 0.0)] = 0.0
        vec **= 2
        # Surrogate + penalty
        return self.penalty * np.asmatrix(np.sum(vec, axis=1))

    def evals(self, xx, d=None, scaling=False):
        """Predict function values

        As a measure of promising function values we let all infeasible points
        have the value of the feasible candidate point with the worst function
        value, since large penalties makes it impossible to distinguish
        between feasible points.

        :param xx: Data points
        :return: Predicted function values
        """
        penalty = self.penalty_fun(xx).T
        vals = self.fhat.evals(xx, d)
        if scaling:
            ind = (np.where(penalty <= 0.0)[0]).T
            if ind.shape[0] > 1:
                ind2 = (np.where(penalty > 0.0)[0]).T
                ind3 = np.argmax(np.squeeze(vals[ind]))
                vals[ind2] = vals[ind3]
                return vals
        return vals + penalty

    def derivs(self, xx, d=None):
        # Compute the value of the constraint functions
        x = np.atleast_2d(xx)
        constraints = np.array(self.data.eval_ineq_constraints(self.from_unit_box(x)))
        dconstraints = self.data.deriv_ineq_constraints(self.from_unit_box(x))
        constraints[np.where(constraints < 0.0)] = 0.0
        return np.atleast_2d(self.fhat.deriv(xx, d)) + \
            2 * self.penalty * np.sum(
                constraints * np.rollaxis(dconstraints, 2), axis=2).T

    def sample_adapt(self):
        """Generate and queue samples from the search strategy"""
        self.adjust_step()
        nsamples = min(self.nsamples, self.maxeval - self.numeval)
        self.search.make_points(self.xbest, self.sigma,
                                self.evals, self.derivs)
        for _ in range(nsamples):
            proposal = self.propose_eval(np.ravel(self.from_unit_box(self.search.next())))
            self.resubmitter.rput(proposal)

    def log_completion(self, record, penalty):
        """Record a completed evaluation to the log.

        :param record: Record of the function evaluation
        :param penalty: Penalty for the given point
        """
        xstr = np.array_str(record.params[0], max_line_width=np.inf,
                            precision=5, suppress_small=True)
        feas = "Feasible"
        if penalty > 0.0:
            feas = "Infeasible"
        logger.info("{} {:.3e} @ {}".format(feas, record.value, xstr))

    def on_complete(self, record):
        """Handle completed function evaluation.

        When a function evaluation is completed we need to ask the constraint
        handler if the function value should be modified which is the case for
        say a penalty method. We also need to print the information to the
        logfile, update the best value found so far and notify the GUI that
        an evaluation has completed.

        :param record: Evaluation record
        """
        x = np.zeros((1, record.params[0].shape[0]))
        x[0, :] = record.params[0]
        penalty = self.penalty_fun(self.to_unit_box(x))[0, 0]
        self.log_completion(record, penalty)
        self.numeval += 1
        record.worker_id = self.worker_id
        record.worker_numeval = self.numeval
        self.fhat.add_point(self.to_unit_box(record.params[0]), record.value)
        # Check if the penalty function is a new best
        if record.value + penalty < self.fbest:
            self.xbest = self.to_unit_box(record.params[0])
            self.fbest = record.value + penalty


from setuptools import setup

setup(
    name='pySOT',
    version='0.1.15',
    packages=['pySOT', 'pySOT.test'],
    url='http://pypi.python.org/pypi/pySOT/',
    license='LICENSE.txt',
    author='David Bindel, David Eriksson',
    author_email='bindel@cornell.edu, dme65@cornell.edu',
    description='Surrogate Optimization Toolbox',
    long_description=open('README.md').read(),	
    install_requires=['pyDOE', 'pyKriging', 'POAP>=0.1.10', 'py_dempster_shafer'],
    classifiers=[
        'Programming Language :: Python :: 2.7',  
    ],
)
